import io
from typing import Dict, Any, Union, Optional, List, Tuple
from typing import TextIO
from warnings import warn

import pandas as pd
from nezha.file import File
from nezha.ustring import clean_str
from pandas import DataFrame
from pandas import Series


def read(file: Union[str, TextIO, io.StringIO],
         names: Optional[Tuple[str, ...]] = None,
         clean_item: bool = True,
         remove_chinese_comma: bool = True,
         skiprows: Union[int, Tuple[int, ...], None] = None,
         header: Union[int, List[int], None] = 0) -> DataFrame:
    """
    read file from xlsx or csv and return pd.DataFrame
    :param file:
    :param names: specified pd.DataFrame header
    :param clean_item: remove invisible char
    :param remove_chinese_comma: not work in xlsx, is suit for csv.
    :header
    指定行数用来作为列名，数据开始行数。
    如果文件中没有列名，则默认为0，否则设置为None。
    如果明确设定header=0 就会替换掉原来存在列名。
    header参数可以是一个list例如：[0,1,3]，这个list表示将文件中的这些行作为列标题（意味着每一列有多个标题），
    介于中间的行将被忽略掉（例如本例中的2；本例中的数据1,2,4行将被作为多级标题出现，第3行数据将被丢弃，dataframe的数据从第5行开始。）。
    :return:
    """
    if isinstance(file, str):
        if file.endswith('xlsx'):
            print(file, names, skiprows, header)
            df = pd.read_excel(file, names=names, skiprows=skiprows, header=header)
        else:
            if not file.endswith('csv'):
                warn(f"{file} treated as csv file")
            if remove_chinese_comma:
                file = File(file).remove_chinese_comma()
            if isinstance(file, io.StringIO):
                df = pd.read_csv(file, names=names, skiprows=skiprows, header=header)
                file.close()
            elif isinstance(file, str):
                df = pd.read_csv(file, names=names, encoding=File(file).encoding, skiprows=skiprows, header=header)
            else:
                raise NotImplementedError(f'file type {type(file)}')
        if not clean_item:
            return df
        frame = []
        for _, series in df.iterrows():
            new_series = {clean_str(k): clean_str(v, striped='"') for k, v in series.iteritems()}
            frame.append(new_series)
        return pd.DataFrame(frame)
    elif isinstance(file, io.StringIO):
        if remove_chinese_comma:
            file = File(file).remove_chinese_comma()
        df = pd.read_csv(file, names=names, skiprows=skiprows, header=header)
        if isinstance(file, io.StringIO):
            file.close()
        return df
    else:
        raise NotImplementedError(f'file type {type(file)}')


def to_excel(df: pd.DataFrame,
             output_name: str,
             index: bool = False,
             columns: Optional[Tuple[str, ...]] = None) -> None:
    df.to_excel(output_name, index=index, columns=columns)


def read_str_csv(s: str, sep: str = '\r\n') -> TextIO:
    """

    :param s:
    example: 'name,phone,idcard\r\n\r\n张**,1536**,14273219***\r\n\r\n韩**,1521**7881,370303**6110320'
    :return: stringIO. csv
    """
    invalid_lines = ('',)
    lines = (line.strip() + '\n' for line in s.split(sep) if line not in invalid_lines)
    io_text = io.StringIO()
    # even call writelines, \n must be added too. It's strange ...
    io_text.writelines(lines)
    io_text.seek(0)
    return io_text


class Excel:

    def __init__(self, file: Union[str, TextIO]):
        """
        as pandas doc, the file is io file-like object.
         you can refer pd.read_html()/pd.read_excel() ... first parameter
        :param file:
        """
        self.file: Union[str, TextIO] = file
        self.dataframe: DataFrame = DataFrame()

    def read(self, names: Optional[Tuple[str, ...]] = None, clean_item: bool = True,
             remove_chinese_comma: bool = True, skiprows: Union[int, Tuple[int, ...], None] = None,
             header: Union[int, List[int], None] = 0) -> 'Excel':
        """

        :param names: if name is not None and the file first row is column_name, skiprows should be 0.
            So the first row will be skipped.
        :param clean_item:
        :param remove_chinese_comma:
        :param skiprows: 需要忽略的行数（从文件开始处算起），或需要跳过的行号列表（从0开始)。
        :return:
        """
        self.dataframe = read(self.file, names=names, clean_item=clean_item, remove_chinese_comma=remove_chinese_comma,
                              skiprows=skiprows, header=header)
        return self

    @property
    def header(self) -> List[str]:
        return list(self.dataframe)

    @classmethod
    def prepare_file(cls, s: str, is_csv: bool = True, sep: str = '\r\n') -> 'Excel':
        if not is_csv:
            raise NotImplementedError()
        file = read_str_csv(s, sep=sep)
        return cls(file)

    def to_excel(self, output_name: str, index: bool = False, columns: Optional[Tuple[str, ...]] = None) -> None:
        print('self.dataframe', self.dataframe)
        to_excel(self.dataframe, output_name, index=index, columns=columns)

    def get_row(self, index: Union[str, int]) -> Series:
        return self.dataframe.loc[index]

    def read_html(self) -> 'Excel':
        self.dataframe = pd.read_html(self.file, encoding='utf-8')
        return self


def csv2xlsx(input: str, output: str, names: Optional[Tuple[str, ...]] = None) -> None:
    """
    usage:
        turn pure csv to xlsx because of scientific notation.
    :param input:
    :param output:
    :param names:
    :return:
    """
    Excel(input).read(names=names).to_excel(output, columns=names)


class Json2Excel:
    """
    template.xlsx content:
    + first row is chinese header
    + second row is json fields

    transform json to excel as the template.xlsx provided
    the transformation rule is:
    + model 'a' matches to json field "a" value
    + model 'a.b' matches to "b" in json field "a" values
    """

    def __init__(self, template_path: str):
        self.template_path: str = template_path
        self._template: Dict = dict()
        self._reversed_template: Dict = dict()

    @classmethod
    def get_recursively(cls, param_key: str, adict: Dict[str, Any]) -> Union[Dict[str, Any], str, None]:
        if isinstance(adict, dict):
            for k, v in adict.items():
                if k != param_key:
                    result = cls.get_recursively(param_key, v)
                    if result is not None:
                        return result
                else:
                    return v
        return None

    @classmethod
    def get_recursively_multi_nest(cls, multi_nest_key: str, data: Dict[str, Any]) -> Any:
        """

        :param multi_nest_key: like 'a.b.c'
        :param data:
        :return:
        """
        fields = multi_nest_key.split('.')
        not_contain_dot = 1
        if len(fields) == not_contain_dot:
            return cls.get_recursively(multi_nest_key, data)
        else:
            first_key, left_keys = fields[0], fields[1:]
            result = cls.get_recursively(first_key, data)
            if isinstance(result, dict):
                nest_key = '.'.join(left_keys)
                return cls.get_recursively_multi_nest(nest_key, result)
            else:
                return result

    def _get_result_val(self, key: str, resp_data: dict) -> str:
        result = self.get_recursively_multi_nest(key, resp_data)
        return result

    @property
    def template(self) -> Dict[str, str]:
        """
        read template.xlsx and transfer to dict.
        :return: dict[chinese, english],
        """
        if not self._template:
            self._template = Excel(self.template_path).read(header=0).dataframe.to_dict(orient='records')[0]
        return self._template

    @property
    def reversed_template(self) -> Dict[str, str]:
        """
        reverse template dict[chinese, english] to dict[chinese, english]
        :return: dict(english=[chinese])
        """
        if not self._reversed_template:
            self._reversed_template = {v: k for k, v in self.template.items()}
        return self._reversed_template

    def parse(self, input: Dict[str, Any], output: str = 'dataframe', index: Optional[List[int]] = None) -> Union[
        Dict[str, Any], pd.DataFrame]:
        """
        parse input by self.reversed_template
        call it like that: result = self.parse(input, output='dataframe', index=index)

        :param input: input data
        :param output: output type: dict/pd.DataFrame
        :param index: if output type is pd.DataFrame and dict value is scalar value, need index
        :return:
        """
        collections = {}
        for key_english, key_chinese in self.reversed_template.items():
            val = self.get_recursively_multi_nest(key_english, input)
            collections[key_chinese] = str(val)
        if output == 'dict':
            return collections
        elif output.lower() == 'dataframe':
            return pd.DataFrame(collections, index=index)
        else:
            raise NotImplementedError(f'output {output} is not support')

    @staticmethod
    def to_excel(df: pd.DataFrame, output: str, columns: Optional[Tuple[str, ...]] = None) -> None:
        """

        :param pf: need persist data
        :param output: output file name
        :param columns: specified the columns name in xlsx file
        :return:
        """
        if not output.endswith('xlsx'):
            raise ValueError(f'output {output} must be xlsx file')
        to_excel(df, output, columns=columns)


if __name__ == '__main__':
    s = '/Users/yutou/Downloads/sqlresult_4210799.csv'
    w = Excel(s).read(names=('name', 'product', 'order_sn', 'price', 'create_at')).to_excel(
        '/Users/yutou/shouxin/sxProject/pysubway/pysubway/gitignore/sqlresut_bankcard4.xlsx',
        columns=('name', 'product', 'order_sn', 'price', 'create_at'), index=False)
    print(w)



import base64
from collections import namedtuple
from typing import Union, Optional, List

from nezha.ustring import to_bytes, to_str

base64_picture = namedtuple('base64_picture', ['format', 'data_str', 'data_binary'])


class Picture:

    @staticmethod
    def to_base64(s: Union[str, bytes, bytearray], is_file: bool = False) -> str:
        if is_file and isinstance(s, str):
            with open(s, mode='rb') as f:
                content = f.read()
        else:
            content = to_bytes(s)
        return to_str(base64.b64encode(content))

    @staticmethod
    def parse_base64(data: Union[str, bytes, bytearray]) -> Optional[base64_picture]:
        """
        usage: get picture format from picture encoded base64

        :param data: picture encoded base64, sample: "image/png;base64,iVBORw0KGgoAAAANSUhEUgAAAh"
        :return: 'jpg' or 'png'; lower letter
        """

        def valid_split(arr: List) -> bool:
            if len(arr) == 1:
                print(sample)
                return False
            return True

        sample = 'valid parameter like: image/png;base64,iVBORw0KGgoAAAANSUhEUgAAAh'
        s: str = to_str(data)

        split: List = s.split(';')
        if not valid_split(split):
            return

        contained_type, contained_data = split
        split_sec: List = contained_type.split('/')
        if not valid_split(split_sec):
            return

        picture_format = 'jpg' if split_sec[1] == 'jpeg' else split_sec[1]
        split_thd: List = contained_data.split(',')
        if not valid_split(split_thd):
            return

        _, picture_data = split_thd
        try:
            data_binary = base64.b64decode(picture_data)
        except Exception as e:
            print(e)
            data_binary = b''
        return base64_picture(picture_format, picture_data, data_binary)



import os

from oss2 import Auth
from oss2 import Bucket


class OSS(object):
    """
    aliyun oss sdk
    """

    def __init__(self, access_key_id: str, access_key_secret: str, bucket_name: str, endpoint: str):
        """

        :param access_key_id: aliyun access_key_id
        :param access_key_secret: aliyun access_key_secret
        :param bucket_name: Bucket名
        :param endpoint: 访问域名或者CNAME
        """
        self.auth = Auth(access_key_id, access_key_secret)
        self.endpoint = endpoint
        self.bucket_name = bucket_name
        self._bucket = None

    @property
    def bucket(self) -> Bucket:
        if not self._bucket:
            endpoint = self.endpoint if self.endpoint.startswith('http') else f'http://{self.endpoint}'
            self._bucket = Bucket(self.auth, endpoint, self.bucket_name)
        return self._bucket

    def sign_url(self, file_path: str, expires=3600 * 24) -> str:
        """
        generate open http url for file in bucket
        :param file_path: 文件的oss存储路径
        """
        return self.bucket.sign_url('GET', file_path, expires)

    @staticmethod
    def generate_file_path(dirpath: str, filename: str) -> str:
        """
        generate upload file_path by filename
        :param filename:
        :return:
        """
        return os.path.join(dirpath, filename)

    def upload(self, file_path: str, file: str) -> bool:
        """
        上传文件到oss
        if picture, file type should be binary.
        :param file_path: 文件的oss存储路径
        :param file: 上传文件
        :return: 上传状态
        """
        result = self.bucket.put_object(key=file_path, data=file)
        return result.status == 200



from collections.abc import Iterable
from typing import Tuple, Union, Mapping

import requests


class Robot:

    def __init__(self, url: Union[str, Tuple[str, ...]]):
        self.url: Tuple[str, ...] = tuple([url, ]) if isinstance(url, str) else url

    def _send(self, data: Mapping) -> None:
        if isinstance(self.url, str):
            raise TypeError(f'data {data} must be iterable and not str')
        if not isinstance(self.url, Iterable):
            raise TypeError(f'data {data} must be iterable and not str')
        for url in self.url:
            requests.post(url, json=data)

    def send_text(self, text: str) -> None:
        data = {"msgtype": "text",
                "text": {
                    "content": f"{text}"
                }
                }
        self._send(data)

    def send_markdown(self, title: str, text: str) -> None:
        data = {
            "msgtype": "markdown",
            "markdown": {
                "title": f"{title}",
                "text": f"{text}"
            }
        }
        self._send(data)


class LogRot(Robot):

    def info(self, text: str) -> None:
        self.send_text(text)

    def error(self, text: str) -> None:
        self.send_text(text)



import json
from typing import Any

import redis
from redis.connection import HiredisParser


class Redis:
    def __init__(self, redis_url: str):
        self.r = redis.from_url(redis_url)

    def set(self, name: str, value: Any, expired_time: int = None) -> HiredisParser:
        """
        设置redis键值对
        :param name: 键名
        :param value: 值
        :param expired_time: 有效期
        """
        if not isinstance(value, str):
            value = json.dumps(value)
        return self.r.set(name, value, expired_time)

    def get(self, name: str) -> Any:
        """
        获取redis键值对
        :param name: 键名
        """
        return self.r.get(name)

    def ttl(self, name: str) -> Any:
        """
        获取redis有效期
        :param name: 键名
        :return: 有效期
        """
        return self.r.ttl(name)

    def delete(self, *name: str) -> HiredisParser:
        """
        删除redis键值对
        :param name: 键名
        """
        return self.r.delete(*name)



import os
import sys
from typing import List, Any, Tuple, Dict

IS_PY3 = sys.version_info.major == 3


def getenv(key: str, default: str, ignore_upper: bool = True) -> str:
    _key = key.lower() if ignore_upper else key
    value = os.getenv(_key) or default
    print("\033[033m", f'{_key} ---------------------> {value}', "\033[0m")
    return value


def filter_locals(variables: Dict[str, Any],
                  exclude_keys: Tuple[str, ...] = ('cls',),
                  exclude_val: Tuple[str, ...] = ('',)) -> List[Tuple[str, Any]]:
    return [(k, v) for k, v in variables.items() if v not in exclude_val and k not in exclude_keys]



# !/usr/bin/env python
# coding: utf-8

from setuptools import find_packages
from setuptools import setup

install_requires = []

setup(
    name='nezha',
    version='0.0.60',
    description="nezha, pysuway's utils",
    packages=find_packages(),
    install_requires=install_requires,
    include_package_data=True
)



import base64
import hashlib
import hmac
from typing import Union, Callable

from nezha.ustring import to_bytes, to_str


def hash_hmac(key: Union[str, bytes],
              plaintext: Union[str, bytes],
              digestmod: Callable = hashlib.sha1,
              returned_str: bool = True,
              returned_base64: bool = False) -> Union[str, bytes]:
    obj = hmac.new(to_bytes(key), to_bytes(plaintext), digestmod=digestmod)
    if returned_base64:
        return to_str(base64.b64encode(obj.digest()))
    if returned_str:
        return obj.hexdigest()
    else:
        return obj.digest()


def hash_md5(plaintext: Union[str, bytes], key: Union[str, bytes] = '') -> str:
    text = to_str(plaintext) + to_str(key)
    m = hashlib.md5()
    m.update(to_bytes(text))
    return m.hexdigest()



from functools import wraps
from typing import Callable, Iterable, Mapping, List, Union, Any


def action_factory(key: str) -> Callable:
    actions = {
        'replace': lambda old, new: new,
        'prefix': lambda old, new: f'{new}{old}',
        'suffix': lambda old, new: f'{old}{new}',
    }
    return actions[key.lower().strip()]


def check_argument(check_func: Callable, arg_flag: Union[int, str],
                   crash_if_failed: bool = True, is_func_decorated: bool = True):
    def _check_argument(arg_flag: Union[int, str], check_func: Callable,
                        crash_if_failed: bool = True, *args, **kwargs) -> Any:
        arg = args[arg_flag] if isinstance(arg_flag, int) else kwargs.get(arg_flag)
        pass_checked = check_func(arg)
        if not pass_checked and crash_if_failed:
            raise ValueError(f'arg {arg} is invalid')
        return pass_checked

    def decorate(func):
        @wraps(func)
        def wrap_method(self, *args, **kwargs):
            _check_argument(arg_flag, check_func, crash_if_failed, *args, **kwargs)
            return func(self, *args, **kwargs)

        @wraps(func)
        def wrap_func(*args, **kwargs):
            _check_argument(arg_flag, check_func, crash_if_failed, *args, **kwargs)
            return func(*args, **kwargs)

        return wrap_func if is_func_decorated else wrap_method

    return decorate


def modify_positional_argument(index: int, val: str, action: str = 'replace', is_func_decorated: bool = True):
    """

    :param index: old positional argument index
    :param val:
    :param action:
        + if action == 'replace', old val -> val
        + if action == 'suffix', old val + val  -> val
        + if action == 'prefix', val + old val -> val
    :param is_func_decorated: if class method or instance method is decorated, self/cls should be passed.
    :return:

    example:

    @modify_positional_argument(0, 'xxxxx', action='prefix')
    def a(xx):
        print(xx)


    class A:

        @modify_positional_argument(0, 'xxxx', is_func_decorated=False)
        def a(self, x):
            print(x)
    """

    def generate_new_args(args: List, index: int, action: str, new_val) -> List:
        new_args = args.copy()
        new_args[index] = action_factory(action)(args[index], new_val)
        return new_args

    def decorate(func):
        @wraps(func)
        def wrap_method(self, *args, **kwargs):
            new_args = generate_new_args(list(args), index, action, val)
            return func(self, *new_args, **kwargs)

        @wraps(func)
        def wrap_func(*args, **kwargs):
            new_args = generate_new_args(list(args), index, action, val)
            return func(*new_args, **kwargs)

        return wrap_func if is_func_decorated else wrap_method

    return decorate


def modify_keyword_argument(new_pairs: Mapping, action: str = 'replace', is_func_decorated: bool = True):
    """

    :param new_pairs: new pair of keyword argument.
    :param action:
        + if action == 'replace', old val -> val
        + if action == 'suffix', old val + val  -> val
        + if action == 'prefix', val + old val -> val
    :param is_func_decorated: if class method or instance method is decorated, self/cls should be passed.
    :return:

    example:

    @modify_keyword_argument({'key': 4444}, action='prefix')
    def a(key=333):
        print(xx)


    class A:

        @modify_keyword_argument({'key': 4444}, action='prefix', is_func_decorated=False)
        def a(self, x):
            print(x)
    """

    def generate_new_pairs(new_pairs, kwargs, action):
        copied_kwargs = kwargs.copy()
        vals = {k: action_factory(action)(kwargs.get(k), new_val) for k, new_val in new_pairs.items()}
        copied_kwargs.update(vals)
        return copied_kwargs

    def decorate(func):
        @wraps(func)
        def wrap_method(self, *args, **kwargs):
            vals = generate_new_pairs(new_pairs, kwargs, action)
            return func(self, *args, **vals)

        @wraps(func)
        def wrap_func(*args, **kwargs):
            vals = generate_new_pairs(new_pairs, kwargs, action)
            return func(*args, **vals)

        return wrap_func if is_func_decorated else wrap_method

    return decorate


def loop(it: Iterable):
    while True:
        try:
            next(it)
        except StopIteration:
            break
